use std::{collections::HashMap, sync::Arc};

use axum::{
    http,
    response::{Html, IntoResponse},
    routing::get,
    Router,
};
use axum_typed_websockets::{Message, WebSocket, WebSocketUpgrade};
use blake2::{Blake2b512, Digest};
use rand::Rng;
use serde::{Deserialize, Serialize};

use offline_web_model::Reference;

#[derive(Debug)]
pub struct AddressableObject {
    pub address: String,
    pub content: String,
}

fn random_object() -> AddressableObject {
    let mut rng = rand::rng();
    let random_size = rng.random_range(50..=4096);
    let random_string: String = (0..random_size)
        .map(|_| rng.sample(rand::distr::Alphanumeric) as char)
        .collect();

    let mut hasher = Blake2b512::new();
    hasher.update(&random_string);
    let hash = format!("{:x}", hasher.finalize());

    AddressableObject {
        address: hash,
        content: random_string,
    }
}

fn random_references_and_objects() -> (
    Arc<Reference>,
    Arc<HashMap<String, Arc<Reference>>>,
    Arc<HashMap<String, AddressableObject>>,
) {
    let path_root = String::from("ref/0");
    let mut objects = HashMap::new();
    let mut refs = HashMap::new();
    let mut root_ref = Reference::new(
        "username:0".to_string(),
        String::from("0"),
        path_root.clone(),
    );
    let mut root_hasher = Blake2b512::new();
    for i in 1..=10 {
        let mut item_ref = Reference::new(
            format!("item:{}", i),
            format!("0:{}", i),
            format!("/item/{}", i),
        );
        let mut hasher = Blake2b512::new();
        for j in 1..=10 {
            let object = random_object();
            hasher.update(&object.content);
            let leaf_ref = Reference::new(
                format!("item:{}:subitem:{}", i, j),
                format!("{}", object.address),
                format!("/item/{}/subitem/{}", i, j),
            )
            .to_arc();
            item_ref = item_ref.add_dep(leaf_ref.clone());
            objects.insert(object.address.clone(), object);
            hasher.update(&leaf_ref.content_address);
            refs.insert(leaf_ref.path.clone(), leaf_ref);
        }
        let hash = format!("{:x}", hasher.finalize());
        item_ref.content_address = hash;
        root_hasher.update(&item_ref.content_address);
        let rc_ref = item_ref.to_arc();
        root_ref = root_ref.add_dep(rc_ref.clone());
        refs.insert(rc_ref.path.clone(), rc_ref);
    }
    root_ref.content_address = format!("{:x}", root_hasher.finalize());
    let rc_root = root_ref.to_arc();
    refs.insert(rc_root.path.clone(), rc_root.clone());
    dbg!(&objects);
    (rc_root, Arc::new(refs), Arc::new(objects))
}

async fn get_client_js() -> impl IntoResponse {
    (
        [(http::header::CONTENT_TYPE, "application/javascript")],
        include_str!("../static/client.js"),
    )
}

#[derive(Debug, Serialize, Deserialize)]
enum ServerMsg {
    Reference(Reference),
    Object(String),
}

#[derive(Debug, Serialize, Deserialize)]
enum ClientMsg {
    Bootstrap,
    GetReference(String),
    GetObject(String),
}

async fn handle_websocket(
    ws: WebSocketUpgrade<ServerMsg, ClientMsg>,
    root_ref: Arc<Reference>,
    refs: Arc<HashMap<String, Arc<Reference>>>,
    objects: Arc<HashMap<String, AddressableObject>>,
) -> impl IntoResponse {
    ws.on_upgrade(|socket| handle_socket(socket, root_ref, refs, objects))
}

async fn handle_socket(
    mut socket: WebSocket<ServerMsg, ClientMsg>,
    root_ref: Arc<Reference>,
    refs: Arc<HashMap<String, Arc<Reference>>>,
    objects: Arc<HashMap<String, AddressableObject>>,
) {
    // Send initial data to the client

    // Handle incoming messages
    // TODO(zaphar): This should actually be smarter in the case of None or Err.
    while let Some(Ok(_msg)) = socket.recv().await {
        match _msg {
            Message::Item(ClientMsg::Bootstrap) => {
                if let Err(e) = socket
                    .send(Message::Item(ServerMsg::Reference(
                        root_ref.as_ref().clone(),
                    )))
                    .await
                {
                    println!("Error sending bootstrap reference: {:?}", e);
                    continue;
                }
            }
            Message::Item(ClientMsg::GetReference(path)) => {
                if let Some(reference) = refs.get(&path) {
                    if let Err(e) = socket
                        .send(Message::Item(ServerMsg::Reference((**reference).clone())))
                        .await
                    {
                        println!("Error sending reference: {:?}", e);
                        continue;
                    }
                } else {
                    println!("Reference not found: {}", path);
                }
            }
            Message::Item(ClientMsg::GetObject(address)) => {
                if let Some(object) = objects.get(&address) {
                    if let Err(e) = socket
                        .send(Message::Item(ServerMsg::Object(object.content.clone())))
                        .await
                    {
                        println!("Error sending object: {:?}", e);
                        continue;
                    }
                } else {
                    println!("Object not found: {}", address);
                }
            }
            Message::Ping(items) => {
                println!("unhandled ping msg: {:?}", items);
            }
            Message::Pong(items) => {
                println!("unhandled pong msg: {:?}", items);
            }
            Message::Close(_close_frame) => {
                println!("closing websocket connection at client request");
                break;
            }
        }
    }

    println!("WebSocket connection closed");
}

pub fn endpoints(
    root_ref: Arc<Reference>,
    refs: Arc<HashMap<String, Arc<Reference>>>,
    objects: Arc<HashMap<String, AddressableObject>>,
) -> Router {
    // WebSocket route now implemented
    Router::new()
        .nest(
            "/api/v1",
            Router::new().route(
                "/ws",
                get(|ws: WebSocketUpgrade<ServerMsg, ClientMsg>| async move {
                    handle_websocket(ws, root_ref.clone(), refs.clone(), objects.clone()).await
                }),
            ),
        )
        .route("/lib/client.js", get(get_client_js))
        .route(
            "/ui/",
            get(|| async { Html(include_str!("../static/index.html")).into_response() }),
        )
}

// TODO(jwall): Javascript test script
pub async fn serve() {
    // run our app with hyper, listening globally on port 3000
    let (root_ref, refs, objects) = random_references_and_objects();
    let listener = tokio::net::TcpListener::bind("127.0.0.1:3000")
        .await
        .unwrap();
    println!("Server ui starting on http://127.0.0.1:3000/ui/");
    println!("WebSocket endpoint available at ws://127.0.0.1:3000/api/v1/ws");
    axum::serve(listener, endpoints(root_ref, refs, objects))
        .await
        .unwrap();
}
